
open Subst;;

(* CCI *)
(* Terms *)

let rec rev_app l1 l2 =
  match l1 with
    [] -> l2
  | x::l -> rev_app l (x::l2)
;;

let rec iter f x = function
    0 -> x
  | n -> iter f (f x) (n-1)
;;

type lambda =
    Rel of int
  | Lam of lambda
  | App of lambda * lambda
  | Constr of int * lambda list
  | Case of lambda * lambda array
;;

let bin_app a b c = App(App(a,b),c);;
let app_list h t = List.fold_left (fun h arg -> App(h,arg)) h t;;
let lam_list n t = iter (fun x -> Lam x) t n;;


let rec lift n k = function
    Rel i -> if i<k then Rel i else Rel (i+n)
  | Lam t -> Lam (lift n (k+1) t)
  | App(a,b) -> App(lift n k a, lift n k b)
  | Constr(i,args) -> Constr(i, List.map (lift n k) args)
  | Case(c,br) -> Case(lift n k c, Array.map (lift n k) br)
;;
let lift n k t = if n = 0 then t else lift n k t

(* Mterms *)

type red_state = Norm | Whnf | Red;;

(* Mutable terms with explicit shifts *)
type mterm =
  { mutable norm: red_state; mutable term: fterm }

 and fterm =
  Frel of int
| Fapp of mterm * mterm
| Flam of mterm * mterm subs * lambda
| Fconstr of int * mterm list
| Fcase of mterm * mterm array
| Flift of int * mterm
| Fclos of mterm subs * lambda
;;

let fapp_list h t =
  List.fold_left (fun h arg -> {norm=Red; term=Fapp(h,arg)}) h t;;
let is_val v = v.norm = Norm;;
let is_whnf v = v.norm <> Red;;
let set_norm v = v.norm <- Norm;;
let set_whnf v = if v.norm = Red then v.norm <- Whnf;; 

let rec lift_fterm n ft =
  match ft.term with
    Flift(k,m) -> lift_fterm (n+k) m
  | _ -> (ft.norm, Flift(n,ft));;

let new_mterm ft = { norm = Red; term = ft };;

(* Lookup in the context *)
let clos_rel e i =
  match expand_rel e i with
      Inl(0,mt) -> mt
    | Inl(n,mt) ->
        let (no,ft) = lift_fterm n mt in { norm=no; term=ft }
    | Inr n -> {norm=Norm; term= Frel n}
;;

(* Optimization: do not enclose variables in a closure.
   Makes variable access much faster *)
let mk_clos e = function
    Rel i -> clos_rel e i
  | t -> new_mterm (Fclos(e,t))
;;

let update v1 (no,t) =
  v1.norm <- no;
  v1.term <- t;
  v1 (* purely functional version: {norm=no; term=t}*)
;;


(* Zipper to map a value of type 'a into a value of type 'b *)
type ('a, 'b) stk =
    Ztop
  | Zappl of 'a * ('a, 'b) stk
  | Zappr of 'b * ('a, 'b) stk
  | Zconstr of int * 'b list * 'a list * ('a, 'b) stk
  | Zcasel of 'a list * ('a, 'b) stk
  | Zcaser of 'b * 'b list * 'a list * ('a, 'b) stk
  | Zlam of mterm subs * lambda * ('a, 'b) stk
  | Zshift of int * ('a, 'b) stk
  | Zupdate of mterm * ('a, 'b) stk
;;

(* Invariants of the stacks:
    - Zshift are collapsed
    - no redex in the stack (e.g. no Zlam(_,_,Zappl(_,_)) )
*)


(* Build a lambda from a mterm, assuming there is no uncomputed closure in it.
   It mainly consists in flushing the shifts. *)
let to_lambda t =
  let rec lbda_rec lfts m stk =
    match m.term with
        Frel i -> zip_lbda (Rel (reloc_rel lfts i)) stk
      | Fapp(a,b) -> lbda_rec lfts a (Zappl ((lfts,b), stk))
      | Flam(u,e,f) -> lbda_rec (llift lfts) u (Zlam (e,f,stk))
      | Fconstr(i,[]) -> zip_lbda (Constr(i,[])) stk
      | Fconstr(i,arg1::args) ->
          lbda_rec lfts arg1
            (Zconstr(i,[],List.map (fun a -> (lfts,a)) args,stk))
      | Fcase(c,br) ->
          lbda_rec lfts c
            (Zcasel(List.map (fun a -> (lfts,a)) (Array.to_list br), stk))
      | Flift (k,m) -> lbda_rec (lshift k lfts) m stk
      | Fclos _ -> failwith "lbda_rec: found a closure"
  and zip_lbda t = function
      Ztop -> t
    | Zappl((lfts,b),s) -> lbda_rec lfts b (Zappr(t,s))
    | Zappr(u,s) -> zip_lbda (App(u,t)) s
    | Zlam(_,_,s) -> zip_lbda (Lam t) s
    | Zconstr(i,l,(lfts,a1)::l',stk) ->
        lbda_rec lfts a1 (Zconstr(i,t::l,l',stk))
    | Zconstr(i,l,[],stk) -> zip_lbda (Constr(i, List.rev (t::l))) stk
    | Zcasel([],stk) -> zip_lbda (Case(t,[||])) stk
    | Zcasel((lfts,b1)::br, stk) -> lbda_rec lfts b1 (Zcaser(t,[],br,stk))
    | Zcaser(c,l,(lfts,b1)::l',stk) ->
        lbda_rec lfts b1 (Zcaser(c,t::l,l',stk))
    | Zcaser(c,l,[],stk) ->
        zip_lbda (Case(c, Array.of_list (List.rev (t::l)))) stk
    | Zshift _ | Zupdate _ -> assert false (* cannot happen *)
  in lbda_rec lid t Ztop
;;

(* Do not generate update marks when it is obviously useless *)
let zup m s =
  if m.norm = Red then Zupdate(m,s) else s
;;

(* Collapse the shifts in the stack *)
let zshift n s =
  match (n,s) with
      (0,_) -> s
    | (_,Zshift(k,s)) -> Zshift(n+k,s)
    | _ -> Zshift(n,s)
;;

(* Tries to rebuild the term until it reaches the top, or find an
   uncomputed closure.
   Preconditions:
    - m must be normal and neutral
    - the argument of Zappr is normal and neutral (untrue in an implementation
      of a call-by-value strategy)
 *)
let rec zip m stk =
  match stk with
    Ztop -> (m, Ztop)
  | Zappr(a,s) -> zip {norm=Norm; term=Fapp(a,m)} s
  | Zconstr(i,l,[],s) -> zip {norm=Norm; term=Fconstr(i,List.rev (m::l))} s
  | Zcasel([],s) -> zip {norm=Norm; term=Fcase(m,[||])} s
  | Zcaser(c,l,[],s) ->
      zip {norm=Norm; term=Fcase(c, Array.of_list (List.rev (m::l)))} s
  | Zlam (e,f,s) -> zip {norm=Norm; term=Flam(m,e,f)} s
  | Zshift(k,s) -> zip {norm=Norm; term=Flift(k,m)} s
  | Zupdate(m',s) -> zip (update m' (Norm, m.term)) s
  | Zappl(a,s) ->
      let s' = Zappr(m,s) in
        if is_val a then zip a s' else (a, s')
  | Zconstr(i,l,a1::l',s) ->
      let s' = Zconstr(i,m::l,l',s) in
        if is_val a1 then zip a1 s' else (a1,s')
  | Zcasel(b1::br,s) ->
      let s' = Zcaser(m,[],br,s) in
        if is_val b1 then zip b1 s' else (b1,s')
  | Zcaser(c,l,a1::l',s) ->
      let s' = Zcaser(c,m::l,l',s) in
        if is_val a1 then zip a1 s' else (a1,s')
;;

(* This "machine" strips the term from applications, lifts and closure.
   Thus, it returns either a value (with tag Norm), or a lambda.
   The result is always a subterm or a reduct of the input. *)
let rec knh m stk =
  if is_val m then (m,stk) else
  match m.term with
  | Fapp(a,b) -> knh a (Zappl (b, zup m stk))
  | Fcase(c,br) -> knh c (Zcasel(Array.to_list br, zup m stk))
  | Flift(k,a) -> knh a (zshift k stk)
  | Frel i -> (set_norm m; (m, stk))
  | Flam (bd,e,f) -> (set_whnf m; (m, stk))
  | Fconstr (i,args) -> (set_whnf m; (m,stk))
  | Fclos(e,f) -> knht e f (Zupdate(m,stk))
 
(* The same for pure terms *)
and knht e t stk =
  match t with
    App(a,b) -> knht e a (Zappl (mk_clos e b, stk))
  | Lam f -> ({norm=Whnf; term=Flam(mk_clos (subs_lift e) f, e, f)}, stk)
  | Rel i -> knh (clos_rel e i) stk
  | Case(c,br) ->
      let brcl = List.map (mk_clos e) (Array.to_list br) in
        knht e c (Zcasel(brcl,stk))
  | Constr(i,args) ->
      ({norm=Whnf; term=Fconstr(i, List.map (mk_clos e) args)}, stk)
;;


(* This "machine" looks for an applied argument in the stack.
   Since the encountered update marks are removed, h must be a whnf *)
let get_arg h stk =
  let rec get_arg_rec depth = function
      Zappl(arg,s) -> (Some(depth,arg), s)
    | Zshift(k,s) -> get_arg_rec (k+depth) s
    | Zupdate(m,s) ->
        update m (lift_fterm depth h);
        get_arg_rec depth s
    | s -> (None, zshift depth s)
  in get_arg_rec 0 stk
;;

let rec is_constr m =
  match m.term with
    Fconstr _ -> true
  | Flift(_,t) -> is_constr t
  | _ -> false
;;

let get_case h stk =
  let rec get_case_rec depth = function
      Zcasel(br,s) -> (Some(depth,br), s)
    | Zshift(k,s) -> get_case_rec (k+depth) s
    | Zupdate(m,s) ->
        update m (lift_fterm depth h);
        get_case_rec depth s
    | s -> (None, zshift depth s)
  in get_case_rec 0 stk
;;

let lift_mterm n t =
  if n=0 then t else
    let (no,nt) = lift_fterm n t in {norm=no; term=nt};;

(* Computes a normal form from the result of knh. *)
let rec knr m stk =
  match m.term with
  | Flam(bd,e,f) ->
      (match get_arg m stk with
           (Some (depth,arg),s) -> knit (subs_shift_cons(depth,e,arg)) f s
         | (None,s) -> if is_val m then (m,s) else kni bd (Zlam (e,f,s)))
  | Fconstr(i,args) ->
      (match get_case m stk with
           (Some (depth,br),s) ->
             kni (fapp_list (List.nth br i)
                    (List.map (lift_mterm depth) args))
               s
         | (None,s) ->
             (match args with
                [] -> (set_norm m; (m,stk))
              | a1::args -> kni a1 (Zconstr(i,[],args,stk))))
  | _ -> (m,stk)

(* Computes the normal form of a term *)
and kni m stk =
  let (hm,s) = knh m stk in
    knr hm s
and knit e t stk =
  let (ht,s) = knht e t stk in
    knr ht s
;;

(* Computes the full normal form of a term.
   1- Calls kni
   2- tries to rebuild the term. If a closure still has to be computed,
      calls itself recursively. *)
let rec walk m stk =
  let (nm,s) = kni m stk in
    match zip nm s with
        (rm,Ztop) -> rm
      | (rm,s') -> walk rm s'
;;

(* Initialization and then normalization *)
let kl t =
  let (m,s) = knit subs_id t Ztop in
    walk m s
;;

let nkni t = to_lambda (kl t)
;;
