
open Terms;;
open Subst;;
open Mterms;;

(********** Lazy **********)
(* An abstract machine based on the closure calculus of Coq V6.2 *)
(* Will probably be integrated in the next release of Coq, but in a more
   functional way! (Qu: is it really an optimization?) *)


(* Zipper to map a value of type 'a into a value of type 'b *)
type ('a, 'b) stk =
    Ztop
  | Zapp of 'b * ('a, 'b) stk
  | Zlam of mterm subs * lambda * ('a, 'b) stk
  | Zlift of int * ('a, 'b) stk
  | Zupdate of mterm * ('a, 'b) stk
  | Zarg of 'a * ('a, 'b) stk
;;

let to_lambda t =
  let rec lbda_rec lfts m stk =
    match m.term with
        Frel i -> zip_lbda (Rel (reloc_rel lfts i)) stk
      | Fapp(a,b) -> lbda_rec lfts a (Zarg ((lfts,b), stk))
      | Flam(u,e,f) -> lbda_rec (llift lfts) u (Zlam (e,f,stk))
      | Flift (k,m) -> lbda_rec (lshift k lfts) m stk
      | Fclos _ -> failwith "lbda_rec"
  and zip_lbda t = function
      Ztop -> t
    | Zapp(u,s) -> zip_lbda (App(u,t)) s
    | Zlam(_,_,s) -> zip_lbda (Lam t) s
    | Zarg((lfts,b),s) -> lbda_rec lfts b (Zapp(t,s))
    | Zlift _ | Zupdate _ -> failwith "zip_lbda"
  in lbda_rec lid t Ztop
;;

let zup m s =
  if m.norm = Red then Zupdate(m,s) else s
;;

let zlift n s =
  match (n,s) with
      (0,_) -> s
    | (_,Zlift(k,s)) -> Zlift(n+k,s)
    | _ -> Zlift(n,s)
;;

(* State of the machines *)
let ge = ref subs_id;;
let gs = ref Ztop;;


(* Tries to rebuild the term until it reaches the top, or find an
   uncomputed closure. *)
let rec zip m =
  match !gs with
    Ztop -> m
  | Zapp(a,s) ->
      gs := s;
      zip {norm=Norm; term=Fapp(a,m)}
  | Zlam (e,f,s) ->
      gs := s;
      zip {norm=Norm; term=Flam(m,e,f)}
  | Zlift(k,s) ->
      gs := s;
      zip {norm=Norm; term=Flift(k,m)}
  | Zupdate(m',s) ->
      gs := s;
      zip (update m' (Norm, m.term))
  | Zarg(a,s) ->
      gs := Zapp(m,s);
      if is_val a then zip a else a
;;

(* This "machine" strips the term from applications, lifts and closure.
   Thus, it returns either a value (with tag Norm), or a lambda. *)
let rec knh m =
  if is_val m then m else
  match m.term with
  | Fapp(a,b) -> (* This part of the code is never used! *)
      gs := Zarg (b, zup m !gs);
      knh a
  | Flift(k,a) ->
      gs := zlift k !gs; (* Update if there are several successive shifts *)
      knh a
  | Frel i ->
      set_norm m;
      m
  | Flam (bd,e,f) ->
      set_whnf m;
      m
  | Fclos(e,f) ->
      ge := e;
      gs := Zupdate(m,!gs);
      knht f

(* The same for pure terms *)
and knht = function
    App(a,b) ->
      gs := Zarg (mk_clos !ge b, !gs);
      knht a
  | Lam f -> {norm=Whnf; term=Flam(mk_clos (subs_lift !ge) f, !ge, f)}
  | Rel i -> knh (clos_rel !ge i)
;;



(* This "machine" looks in the stack if there is an applied argument *)
let get_arg h =
  let rec get_arg_rec depth = function
      Zarg(arg,s) ->
        gs := s;
        Some(depth,arg)
    | Zlift(k,s) -> get_arg_rec (k+depth) s
    | Zupdate(m,s) ->
        update m (lift_fterm depth h);
        get_arg_rec depth s
    | s ->
        gs := zlift depth s;
        None
  in get_arg_rec 0 !gs
;;

(* Computes a normal form from the result of knh. *)
let rec knr m =
  match m.term with
  | Flam (bd,e,f) ->
      (match get_arg m with
           Some (depth,arg) -> knit (subs_shift_cons(depth,e,arg)) f
         | None ->
             if is_val m then m else
               (gs := Zlam (e,f,!gs);
                kni bd))
  | _ -> m

(* Computes the normal form of a term *)
and kni m = knr (knh m)
and knit e t =
  ge := e;
  knr (knht t)
;;

(* Computes the full normal form of a term.
   1- Calls kni
   2- tries to rebuild the term. If a closure still has to be conputed,
      calls itself recursively. *)
let rec walk m =
  let rm = zip (kni m) in
    match !gs with
        Ztop -> rm
      | _ -> walk rm
;;

(* Initialization and then normalization *)
let kl t =
  gs := Ztop;
  walk (knit subs_id t)
;;


let nkni t = to_lambda (kl t)
;;
